###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import hashlib
import os
import threading
from typing import Any, Dict, Optional, Union

from autobahn.wamp.types import TransportDetails
from twisted.internet.address import IPv4Address, UNIXAddress
from twisted.internet.defer import Deferred
from twisted.internet.interfaces import IProcessTransport, ITransport

try:
    from twisted.internet.stdio import PipeAddress
except ImportError:
    # stdio.PipeAddress is only avail on Twisted 13.0+
    PipeAddress = type(None)

try:
    from twisted.internet.address import IPv6Address

    _HAS_IPV6 = True
except ImportError:
    _HAS_IPV6 = False
    IPv6Address = type(None)

try:
    from OpenSSL.SSL import Connection

    from twisted.internet.interfaces import ISSLTransport
    from twisted.protocols.tls import TLSMemoryBIOProtocol

    _HAS_TLS = True
except ImportError:
    _HAS_TLS = False

__all = (
    "sleep",
    "peer2str",
    "transport_channel_id",
    "extract_peer_certificate",
    "create_transport_details",
)


def sleep(delay, reactor=None):
    """
    Inline sleep for use in co-routines (Twisted ``inlineCallback`` decorated functions).

    .. seealso::
       * `twisted.internet.defer.inlineCallbacks <http://twistedmatrix.com/documents/current/api/twisted.internet.defer.html#inlineCallbacks>`__
       * `twisted.internet.interfaces.IReactorTime <http://twistedmatrix.com/documents/current/api/twisted.internet.interfaces.IReactorTime.html>`__

    :param delay: Time to sleep in seconds.
    :type delay: float
    :param reactor: The Twisted reactor to use.
    :type reactor: None or provider of ``IReactorTime``.
    """
    if not reactor:
        from twisted.internet import reactor
    d = Deferred()
    reactor.callLater(delay, d.callback, None)
    return d


def peer2str(transport: Union[ITransport, IProcessTransport]) -> str:
    """
    Return a *peer descriptor* given a Twisted transport, for example:

    * ``tcp4:127.0.0.1:52914``: a TCPv4 socket
    * ``unix:/tmp/server.sock``: a Unix domain socket
    * ``process:142092``: a Pipe originating from a spawning (parent) process
    * ``pipe``: a Pipe terminating in a spawned (child) process

    :returns: Returns a string representation of the peer of the Twisted transport.
    """
    # IMPORTANT: we need to _first_ test for IProcessTransport
    if IProcessTransport.providedBy(transport):
        # note the PID of the forked process in the peer descriptor
        res = "process:{}".format(transport.pid)
    elif ITransport.providedBy(transport):
        addr: Union[IPv4Address, IPv6Address, UNIXAddress, PipeAddress] = (
            transport.getPeer()
        )
        if isinstance(addr, IPv4Address):
            res = "tcp4:{0}:{1}".format(addr.host, addr.port)
        elif _HAS_IPV6 and isinstance(addr, IPv6Address):
            res = "tcp6:{0}:{1}".format(addr.host, addr.port)
        elif isinstance(addr, UNIXAddress):
            if addr.name:
                res = "unix:{0}".format(addr.name)
            else:
                res = "unix"
        elif isinstance(addr, PipeAddress):
            # sadly, we don't have a way to get at the PID of the other side of the pipe
            # res = "pipe"
            res = "process:{0}".format(os.getppid())
        else:
            # gracefully fallback if we can't map the peer's address
            res = "unknown"
    else:
        # gracefully fallback if we can't map the peer's transport
        res = "unknown"
    return res


if not _HAS_TLS:

    def transport_channel_id(
        transport: object, is_server: bool, channel_id_type: Optional[str] = None
    ) -> Optional[bytes]:
        if channel_id_type is None:
            return b"\x00" * 32
        else:
            raise RuntimeError(
                'cannot determine TLS channel ID of type "{}" when TLS is not available on this system'.format(
                    channel_id_type
                )
            )

else:

    def transport_channel_id(
        transport: object, is_server: bool, channel_id_type: Optional[str] = None
    ) -> Optional[bytes]:
        """
        Return TLS channel ID of WAMP transport of the given TLS channel ID type.

        Application-layer user authentication protocols are vulnerable to generic credential forwarding attacks,
        where an authentication credential sent by a client C to a server M may then be used by M to impersonate C at
        another server S.
        To prevent such credential forwarding attacks, modern authentication protocols rely on channel bindings.
        For example, WAMP-cryptosign can use the tls-unique channel identifier provided by the TLS layer to strongly
        bind authentication credentials to the underlying channel, so that a credential received on one TLS channel
        cannot be forwarded on another.

        :param transport: The Twisted TLS transport to extract the TLS channel ID from. If the transport isn't
            TLS based, and non-empty ``channel_id_type`` is requested, ``None`` will be returned. If the transport
            is indeed TLS based, an empty ``channel_id_type`` of ``None`` is requested, 32 NUL bytes will be returned.
        :param is_server: Flag indicating that the transport is a server transport.
        :param channel_id_type: TLS channel ID type, if set currently only ``"tls-unique"`` is supported.
        :returns: The TLS channel ID (32 bytes).
        """
        if channel_id_type is None:
            return b"\x00" * 32

        if channel_id_type not in ["tls-unique"]:
            raise RuntimeError(
                'invalid TLS channel ID type "{}" requested'.format(channel_id_type)
            )

        if not isinstance(transport, TLSMemoryBIOProtocol):
            raise RuntimeError(
                'cannot determine TLS channel ID of type "{}" when TLS is not available on this transport {}'.format(
                    channel_id_type, type(transport)
                )
            )

        # get access to the OpenSSL connection underlying the Twisted protocol
        # https://twistedmatrix.com/documents/current/api/twisted.protocols.tls.TLSMemoryBIOProtocol.html#getHandle
        connection: Connection = transport.getHandle()
        assert connection and isinstance(connection, Connection)

        # Obtain latest TLS Finished message that we expected from peer, or None if handshake is not completed.
        # http://www.pyopenssl.org/en/stable/api/ssl.html#OpenSSL.SSL.Connection.get_peer_finished
        is_not_resumed = True

        if channel_id_type == "tls-unique":
            # see also: https://bugs.python.org/file22646/tls_channel_binding.patch
            if is_server != is_not_resumed:
                # for routers (=servers) XOR new sessions, the channel ID is based on the TLS Finished message we
                # expected to receive from the client: contents of the message or None if the TLS handshake has
                # not yet completed.
                tls_finished_msg = connection.get_peer_finished()
            else:
                # for clients XOR resumed sessions, the channel ID is based on the TLS Finished message we sent
                # to the router (=server): contents of the message or None if the TLS handshake has not yet completed.
                tls_finished_msg = connection.get_finished()

            if tls_finished_msg is None:
                # this can occur when:
                #   1. we made a successful connection (in a TCP sense) but something failed with
                #      the TLS handshake (e.g. invalid certificate)
                #   2. the TLS handshake has not yet completed
                return b"\x00" * 32
            else:
                m = hashlib.sha256()
                m.update(tls_finished_msg)
                return m.digest()
        else:
            raise NotImplementedError(
                'should not arrive here (unhandled channel_id_type "{}")'.format(
                    channel_id_type
                )
            )


if not _HAS_TLS:

    def extract_peer_certificate(transport: object) -> Optional[Dict[str, Any]]:
        """
        Dummy when no TLS is available.

        :param transport: Ignored.
        :return: Always return ``None``.
        """
        return None

else:

    def extract_peer_certificate(
        transport: TLSMemoryBIOProtocol,
    ) -> Optional[Dict[str, Any]]:
        """
        Extract TLS x509 client certificate information from a Twisted stream transport, and
        return a dict with x509 TLS client certificate information (if the client provided a
        TLS client certificate).

        :param transport: The secure transport from which to extract the peer certificate (if present).
        :returns: If the peer provided a certificate, the parsed certificate information set.
        """
        # check if the Twisted transport is a TLSMemoryBIOProtocol
        if not (
            ISSLTransport.providedBy(transport)
            and hasattr(transport, "getPeerCertificate")
        ):
            return None

        cert = transport.getPeerCertificate()
        if cert:
            # extract x509 name components from an OpenSSL X509Name object
            def maybe_bytes(_value):
                if isinstance(_value, bytes):
                    return _value.decode("utf8")
                else:
                    return _value

            result = {
                "md5": "{}".format(maybe_bytes(cert.digest("md5"))).upper(),
                "sha1": "{}".format(maybe_bytes(cert.digest("sha1"))).upper(),
                "sha256": "{}".format(maybe_bytes(cert.digest("sha256"))).upper(),
                "expired": bool(cert.has_expired()),
                "hash": maybe_bytes(cert.subject_name_hash()),
                "serial": int(cert.get_serial_number()),
                "signature_algorithm": maybe_bytes(cert.get_signature_algorithm()),
                "version": int(cert.get_version()),
                "not_before": maybe_bytes(cert.get_notBefore()),
                "not_after": maybe_bytes(cert.get_notAfter()),
                "extensions": [],
            }

            for i in range(cert.get_extension_count()):
                ext = cert.get_extension(i)
                ext_info = {
                    "name": "{}".format(maybe_bytes(ext.get_short_name())),
                    "value": "{}".format(maybe_bytes(ext)),
                    "critical": ext.get_critical() != 0,
                }
                result["extensions"].append(ext_info)

            for entity, name in [
                ("subject", cert.get_subject()),
                ("issuer", cert.get_issuer()),
            ]:
                result[entity] = {}
                for key, value in name.get_components():
                    key = maybe_bytes(key)
                    value = maybe_bytes(value)
                    result[entity]["{}".format(key).lower()] = "{}".format(value)

            return result


def create_transport_details(
    transport: Union[ITransport, IProcessTransport], is_server: bool
) -> TransportDetails:
    """
    Create transport details from Twisted transport.

    :param transport: The Twisted transport to extract information from.
    :param is_server: Flag indicating whether this transport side is a "server" (as in TCP server).
    :return: Transport details object filled with information from the Twisted transport.
    """
    peer = peer2str(transport)

    own_pid = os.getpid()
    if hasattr(threading, "get_native_id"):
        # New in Python 3.8
        # https://docs.python.org/3/library/threading.html?highlight=get_native_id#threading.get_native_id
        own_tid = threading.get_native_id()
    else:
        own_tid = threading.get_ident()
    own_fd = -1

    if _HAS_TLS and ISSLTransport.providedBy(transport):
        channel_id = {
            # this will only be filled when the TLS opening handshake is complete (!)
            "tls-unique": transport_channel_id(transport, is_server, "tls-unique"),
        }
        channel_type = TransportDetails.CHANNEL_TYPE_TLS
        peer_cert = extract_peer_certificate(transport)
        is_secure = True
    else:
        channel_id = {}
        channel_type = TransportDetails.CHANNEL_TYPE_TCP
        peer_cert = None
        is_secure = False

    # FIXME: really set a default (websocket)?
    channel_framing = TransportDetails.CHANNEL_FRAMING_WEBSOCKET

    td = TransportDetails(
        channel_type=channel_type,
        channel_framing=channel_framing,
        peer=peer,
        is_server=is_server,
        own_pid=own_pid,
        own_tid=own_tid,
        own_fd=own_fd,
        is_secure=is_secure,
        channel_id=channel_id,
        peer_cert=peer_cert,
    )

    return td



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: reflection

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class Schema(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = Schema()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsSchema(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    @classmethod
    def SchemaBufferHasIdentifier(cls, buf, offset, size_prefixed=False):
        return flatbuffers.util.BufferHasIdentifier(
            buf, offset, b"\x42\x46\x42\x53", size_prefixed=size_prefixed
        )

    # Schema
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # Schema
    def Objects(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from reflection.Object import Object

            obj = Object()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Schema
    def ObjectsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Schema
    def ObjectsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        return o == 0

    # Schema
    def Enums(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from reflection.Enum import Enum

            obj = Enum()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Schema
    def EnumsLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Schema
    def EnumsIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        return o == 0

    # Schema
    def FileIdent(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Schema
    def FileExt(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # Schema
    def RootTable(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            x = self._tab.Indirect(o + self._tab.Pos)
            from reflection.Object import Object

            obj = Object()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Schema
    def Services(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from reflection.Service import Service

            obj = Service()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Schema
    def ServicesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Schema
    def ServicesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        return o == 0

    # Schema
    def AdvancedFeatures(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint64Flags, o + self._tab.Pos
            )
        return 0

    # All the files used in this compilation. Files are relative to where
    # flatc was invoked.
    # Schema
    def FbsFiles(self, j):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            x = self._tab.Vector(o)
            x += flatbuffers.number_types.UOffsetTFlags.py_type(j) * 4
            x = self._tab.Indirect(x)
            from reflection.SchemaFile import SchemaFile

            obj = SchemaFile()
            obj.Init(self._tab.Bytes, x)
            return obj
        return None

    # Schema
    def FbsFilesLength(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return self._tab.VectorLen(o)
        return 0

    # Schema
    def FbsFilesIsNone(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        return o == 0


def SchemaStart(builder):
    builder.StartObject(8)


def Start(builder):
    SchemaStart(builder)


def SchemaAddObjects(builder, objects):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(objects), 0
    )


def AddObjects(builder, objects):
    SchemaAddObjects(builder, objects)


def SchemaStartObjectsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartObjectsVector(builder, numElems):
    return SchemaStartObjectsVector(builder, numElems)


def SchemaAddEnums(builder, enums):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(enums), 0
    )


def AddEnums(builder, enums):
    SchemaAddEnums(builder, enums)


def SchemaStartEnumsVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartEnumsVector(builder, numElems):
    return SchemaStartEnumsVector(builder, numElems)


def SchemaAddFileIdent(builder, fileIdent):
    builder.PrependUOffsetTRelativeSlot(
        2, flatbuffers.number_types.UOffsetTFlags.py_type(fileIdent), 0
    )


def AddFileIdent(builder, fileIdent):
    SchemaAddFileIdent(builder, fileIdent)


def SchemaAddFileExt(builder, fileExt):
    builder.PrependUOffsetTRelativeSlot(
        3, flatbuffers.number_types.UOffsetTFlags.py_type(fileExt), 0
    )


def AddFileExt(builder, fileExt):
    SchemaAddFileExt(builder, fileExt)


def SchemaAddRootTable(builder, rootTable):
    builder.PrependUOffsetTRelativeSlot(
        4, flatbuffers.number_types.UOffsetTFlags.py_type(rootTable), 0
    )


def AddRootTable(builder, rootTable):
    SchemaAddRootTable(builder, rootTable)


def SchemaAddServices(builder, services):
    builder.PrependUOffsetTRelativeSlot(
        5, flatbuffers.number_types.UOffsetTFlags.py_type(services), 0
    )


def AddServices(builder, services):
    SchemaAddServices(builder, services)


def SchemaStartServicesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartServicesVector(builder, numElems):
    return SchemaStartServicesVector(builder, numElems)


def SchemaAddAdvancedFeatures(builder, advancedFeatures):
    builder.PrependUint64Slot(6, advancedFeatures, 0)


def AddAdvancedFeatures(builder, advancedFeatures):
    SchemaAddAdvancedFeatures(builder, advancedFeatures)


def SchemaAddFbsFiles(builder, fbsFiles):
    builder.PrependUOffsetTRelativeSlot(
        7, flatbuffers.number_types.UOffsetTFlags.py_type(fbsFiles), 0
    )


def AddFbsFiles(builder, fbsFiles):
    SchemaAddFbsFiles(builder, fbsFiles)


def SchemaStartFbsFilesVector(builder, numElems):
    return builder.StartVector(4, numElems, 4)


def StartFbsFilesVector(builder, numElems):
    return SchemaStartFbsFilesVector(builder, numElems)


def SchemaEnd(builder):
    return builder.EndObject()


def End(builder):
    return SchemaEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class AuthCraChallenge(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = AuthCraChallenge()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsAuthCraChallenge(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # AuthCraChallenge
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # AuthCraChallenge
    def Challenge(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # AuthCraChallenge
    def Salt(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return self._tab.String(o + self._tab.Pos)
        return None

    # AuthCraChallenge
    def Iterations(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return self._tab.Get(
                flatbuffers.number_types.Uint32Flags, o + self._tab.Pos
            )
        return 1000

    # AuthCraChallenge
    def Keylen(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return self._tab.Get(flatbuffers.number_types.Uint8Flags, o + self._tab.Pos)
        return 32


def AuthCraChallengeStart(builder):
    builder.StartObject(4)


def Start(builder):
    AuthCraChallengeStart(builder)


def AuthCraChallengeAddChallenge(builder, challenge):
    builder.PrependUOffsetTRelativeSlot(
        0, flatbuffers.number_types.UOffsetTFlags.py_type(challenge), 0
    )


def AddChallenge(builder, challenge):
    AuthCraChallengeAddChallenge(builder, challenge)


def AuthCraChallengeAddSalt(builder, salt):
    builder.PrependUOffsetTRelativeSlot(
        1, flatbuffers.number_types.UOffsetTFlags.py_type(salt), 0
    )


def AddSalt(builder, salt):
    AuthCraChallengeAddSalt(builder, salt)


def AuthCraChallengeAddIterations(builder, iterations):
    builder.PrependUint32Slot(2, iterations, 1000)


def AddIterations(builder, iterations):
    AuthCraChallengeAddIterations(builder, iterations)


def AuthCraChallengeAddKeylen(builder, keylen):
    builder.PrependUint8Slot(3, keylen, 32)


def AddKeylen(builder, keylen):
    AuthCraChallengeAddKeylen(builder, keylen)


def AuthCraChallengeEnd(builder):
    return builder.EndObject()


def End(builder):
    return AuthCraChallengeEnd(builder)



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class BrokerFeatures(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = BrokerFeatures()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsBrokerFeatures(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # BrokerFeatures
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # BrokerFeatures
    def PublisherIdentification(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # BrokerFeatures
    def PublisherExclusion(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # BrokerFeatures
    def SubscriberBlackwhiteListing(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # BrokerFeatures
    def PatternBasedSubscription(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # BrokerFeatures
    def PublicationTrustlevels(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # BrokerFeatures
    def SubscriptionRevocation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # BrokerFeatures
    def SessionMetaApi(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # BrokerFeatures
    def SubscriptionMetaApi(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # BrokerFeatures
    def EventRetention(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(20))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # BrokerFeatures
    def EventHistory(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(22))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # BrokerFeatures
    def AcknowledgeEventReceived(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(24))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # BrokerFeatures
    def AcknowledgeSubscriberReceived(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(26))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # BrokerFeatures
    def PayloadTransparency(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(28))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # BrokerFeatures
    def PayloadEncryptionCryptobox(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(30))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False


def BrokerFeaturesStart(builder):
    builder.StartObject(14)


def Start(builder):
    BrokerFeaturesStart(builder)


def BrokerFeaturesAddPublisherIdentification(builder, publisherIdentification):
    builder.PrependBoolSlot(0, publisherIdentification, 0)


def AddPublisherIdentification(builder, publisherIdentification):
    BrokerFeaturesAddPublisherIdentification(builder, publisherIdentification)


def BrokerFeaturesAddPublisherExclusion(builder, publisherExclusion):
    builder.PrependBoolSlot(1, publisherExclusion, 0)


def AddPublisherExclusion(builder, publisherExclusion):
    BrokerFeaturesAddPublisherExclusion(builder, publisherExclusion)


def BrokerFeaturesAddSubscriberBlackwhiteListing(builder, subscriberBlackwhiteListing):
    builder.PrependBoolSlot(2, subscriberBlackwhiteListing, 0)


def AddSubscriberBlackwhiteListing(builder, subscriberBlackwhiteListing):
    BrokerFeaturesAddSubscriberBlackwhiteListing(builder, subscriberBlackwhiteListing)


def BrokerFeaturesAddPatternBasedSubscription(builder, patternBasedSubscription):
    builder.PrependBoolSlot(3, patternBasedSubscription, 0)


def AddPatternBasedSubscription(builder, patternBasedSubscription):
    BrokerFeaturesAddPatternBasedSubscription(builder, patternBasedSubscription)


def BrokerFeaturesAddPublicationTrustlevels(builder, publicationTrustlevels):
    builder.PrependBoolSlot(4, publicationTrustlevels, 0)


def AddPublicationTrustlevels(builder, publicationTrustlevels):
    BrokerFeaturesAddPublicationTrustlevels(builder, publicationTrustlevels)


def BrokerFeaturesAddSubscriptionRevocation(builder, subscriptionRevocation):
    builder.PrependBoolSlot(5, subscriptionRevocation, 0)


def AddSubscriptionRevocation(builder, subscriptionRevocation):
    BrokerFeaturesAddSubscriptionRevocation(builder, subscriptionRevocation)


def BrokerFeaturesAddSessionMetaApi(builder, sessionMetaApi):
    builder.PrependBoolSlot(6, sessionMetaApi, 0)


def AddSessionMetaApi(builder, sessionMetaApi):
    BrokerFeaturesAddSessionMetaApi(builder, sessionMetaApi)


def BrokerFeaturesAddSubscriptionMetaApi(builder, subscriptionMetaApi):
    builder.PrependBoolSlot(7, subscriptionMetaApi, 0)


def AddSubscriptionMetaApi(builder, subscriptionMetaApi):
    BrokerFeaturesAddSubscriptionMetaApi(builder, subscriptionMetaApi)


def BrokerFeaturesAddEventRetention(builder, eventRetention):
    builder.PrependBoolSlot(8, eventRetention, 0)


def AddEventRetention(builder, eventRetention):
    BrokerFeaturesAddEventRetention(builder, eventRetention)


def BrokerFeaturesAddEventHistory(builder, eventHistory):
    builder.PrependBoolSlot(9, eventHistory, 0)


def AddEventHistory(builder, eventHistory):
    BrokerFeaturesAddEventHistory(builder, eventHistory)


def BrokerFeaturesAddAcknowledgeEventReceived(builder, acknowledgeEventReceived):
    builder.PrependBoolSlot(10, acknowledgeEventReceived, 0)


def AddAcknowledgeEventReceived(builder, acknowledgeEventReceived):
    BrokerFeaturesAddAcknowledgeEventReceived(builder, acknowledgeEventReceived)


def BrokerFeaturesAddAcknowledgeSubscriberReceived(
    builder, acknowledgeSubscriberReceived
):
    builder.PrependBoolSlot(11, acknowledgeSubscriberReceived, 0)


def AddAcknowledgeSubscriberReceived(builder, acknowledgeSubscriberReceived):
    BrokerFeaturesAddAcknowledgeSubscriberReceived(
        builder, acknowledgeSubscriberReceived
    )


def BrokerFeaturesAddPayloadTransparency(builder, payloadTransparency):
    builder.PrependBoolSlot(12, payloadTransparency, 0)


def AddPayloadTransparency(builder, payloadTransparency):
    BrokerFeaturesAddPayloadTransparency(builder, payloadTransparency)


def BrokerFeaturesAddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox):
    builder.PrependBoolSlot(13, payloadEncryptionCryptobox, 0)


def AddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox):
    BrokerFeaturesAddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox)


def BrokerFeaturesEnd(builder):
    return builder.EndObject()


def End(builder):
    return BrokerFeaturesEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

# Import brotli - will be either 'brotli' (CPython) or 'brotlicffi' (PyPy)
# The actual import is handled in compress.py with platform detection
try:
    import brotli
except ImportError:
    import brotlicffi as brotli

from autobahn.websocket.compress_base import (
    PerMessageCompress,
    PerMessageCompressOffer,
    PerMessageCompressOfferAccept,
    PerMessageCompressResponse,
    PerMessageCompressResponseAccept,
)

__all__ = (
    "PerMessageBrotli",
    "PerMessageBrotliMixin",
    "PerMessageBrotliOffer",
    "PerMessageBrotliOfferAccept",
    "PerMessageBrotliResponse",
    "PerMessageBrotliResponseAccept",
)


class PerMessageBrotliMixin(object):
    """
    Mixin class for this extension.
    """

    EXTENSION_NAME = "permessage-brotli"
    """
    Name of this WebSocket extension.
    """


class PerMessageBrotliOffer(PerMessageCompressOffer, PerMessageBrotliMixin):
    """
    Set of extension parameters for `permessage-brotli` WebSocket extension
    offered by a client to a server.
    """

    @classmethod
    def parse(cls, params):
        """
        Parses a WebSocket extension offer for `permessage-brotli` provided by a client to a server.

        :param params: Output from :func:`autobahn.websocket.WebSocketProtocol._parseExtensionsHeader`.
        :type params: list

        :returns: A new instance of :class:`autobahn.compress.PerMessageBrotliOffer`.
        :rtype: obj
        """
        # extension parameter defaults
        accept_no_context_takeover = False
        request_no_context_takeover = False

        # verify/parse client ("client-to-server direction") parameters of permessage-brotli offer
        for p in params:
            if len(params[p]) > 1:
                raise Exception(
                    "multiple occurrence of extension parameter '%s' for extension '%s'"
                    % (p, cls.EXTENSION_NAME)
                )

            val = params[p][0]

            if p == "client_no_context_takeover":
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception(
                        "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                        % (val, p, cls.EXTENSION_NAME)
                    )
                else:
                    accept_no_context_takeover = True

            elif p == "server_no_context_takeover":
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception(
                        "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                        % (val, p, cls.EXTENSION_NAME)
                    )
                else:
                    request_no_context_takeover = True

            else:
                raise Exception(
                    "illegal extension parameter '%s' for extension '%s'"
                    % (p, cls.EXTENSION_NAME)
                )

        offer = cls(accept_no_context_takeover, request_no_context_takeover)
        return offer

    def __init__(
        self, accept_no_context_takeover=True, request_no_context_takeover=False
    ):
        """

        :param accept_no_context_takeover: Iff true, client accepts "no context takeover" feature.
        :type accept_no_context_takeover: bool
        :param request_no_context_takeover: Iff true, client request "no context takeover" feature.
        :type request_no_context_takeover: bool
        """
        if type(accept_no_context_takeover) != bool:
            raise Exception(
                "invalid type %s for accept_no_context_takeover"
                % type(accept_no_context_takeover)
            )

        self.accept_no_context_takeover = accept_no_context_takeover

        if type(request_no_context_takeover) != bool:
            raise Exception(
                "invalid type %s for request_no_context_takeover"
                % type(request_no_context_takeover)
            )

        self.request_no_context_takeover = request_no_context_takeover

    def get_extension_string(self):
        """
        Returns the WebSocket extension configuration string as sent to the server.

        :returns: PMCE configuration string.
        :rtype: str
        """
        pmce_string = self.EXTENSION_NAME
        if self.accept_no_context_takeover:
            pmce_string += "; client_no_context_takeover"
        if self.request_no_context_takeover:
            pmce_string += "; server_no_context_takeover"
        return pmce_string

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: JSON serializable representation.
        :rtype: dict
        """
        return {
            "extension": self.EXTENSION_NAME,
            "accept_no_context_takeover": self.accept_no_context_takeover,
            "request_no_context_takeover": self.request_no_context_takeover,
        }

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: Python string representation.
        :rtype: str
        """
        return (
            "PerMessageBrotliOffer(accept_no_context_takeover = %s, request_no_context_takeover = %s)"
            % (self.accept_no_context_takeover, self.request_no_context_takeover)
        )


class PerMessageBrotliOfferAccept(PerMessageCompressOfferAccept, PerMessageBrotliMixin):
    """
    Set of parameters with which to accept an `permessage-brotli` offer
    from a client by a server.
    """

    def __init__(
        self, offer, request_no_context_takeover=False, no_context_takeover=None
    ):
        """

        :param offer: The offer being accepted.
        :type offer: Instance of :class:`autobahn.compress.PerMessageBrotliOffer`.
        :param request_no_context_takeover: Iff true, server request "no context takeover" feature.
        :type request_no_context_takeover: bool
        :param no_context_takeover: Override server ("server-to-client direction") context takeover (this must be compatible with offer).
        :type no_context_takeover: bool
        """
        if not isinstance(offer, PerMessageBrotliOffer):
            raise Exception("invalid type %s for offer" % type(offer))

        self.offer = offer

        if type(request_no_context_takeover) != bool:
            raise Exception(
                "invalid type %s for request_no_context_takeover"
                % type(request_no_context_takeover)
            )

        if request_no_context_takeover and not offer.accept_no_context_takeover:
            raise Exception(
                "invalid value %s for request_no_context_takeover - feature unsupported by client"
                % request_no_context_takeover
            )

        self.request_no_context_takeover = request_no_context_takeover

        if no_context_takeover is not None:
            if type(no_context_takeover) != bool:
                raise Exception(
                    "invalid type %s for no_context_takeover"
                    % type(no_context_takeover)
                )

            if offer.request_no_context_takeover and not no_context_takeover:
                raise Exception(
                    "invalid value %s for no_context_takeover - client requested feature"
                    % no_context_takeover
                )

        self.no_context_takeover = no_context_takeover

    def get_extension_string(self):
        """
        Returns the WebSocket extension configuration string as sent to the server.

        :returns: PMCE configuration string.
        :rtype: str
        """
        pmce_string = self.EXTENSION_NAME
        if self.offer.request_no_context_takeover:
            pmce_string += "; server_no_context_takeover"
        if self.request_no_context_takeover:
            pmce_string += "; client_no_context_takeover"
        return pmce_string

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: JSON serializable representation.
        :rtype: dict
        """
        return {
            "extension": self.EXTENSION_NAME,
            "offer": self.offer.__json__(),
            "request_no_context_takeover": self.request_no_context_takeover,
            "no_context_takeover": self.no_context_takeover,
        }

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: Python string representation.
        :rtype: str
        """
        return (
            "PerMessageBrotliAccept(offer = %s, request_no_context_takeover = %s, no_context_takeover = %s)"
            % (
                self.offer.__repr__(),
                self.request_no_context_takeover,
                self.no_context_takeover,
            )
        )


class PerMessageBrotliResponse(PerMessageCompressResponse, PerMessageBrotliMixin):
    """
    Set of parameters for `permessage-brotli` responded by server.
    """

    @classmethod
    def parse(cls, params):
        """
        Parses a WebSocket extension response for `permessage-brotli` provided by a server to a client.

        :param params: Output from :func:`autobahn.websocket.WebSocketProtocol._parseExtensionsHeader`.
        :type params: list

        :returns: A new instance of :class:`autobahn.compress.PerMessageBrotliResponse`.
        :rtype: obj
        """
        client_no_context_takeover = False
        server_no_context_takeover = False

        for p in params:
            if len(params[p]) > 1:
                raise Exception(
                    "multiple occurrence of extension parameter '%s' for extension '%s'"
                    % (p, cls.EXTENSION_NAME)
                )

            val = params[p][0]

            if p == "client_no_context_takeover":
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception(
                        "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                        % (val, p, cls.EXTENSION_NAME)
                    )
                else:
                    client_no_context_takeover = True

            elif p == "server_no_context_takeover":
                # noinspection PySimplifyBooleanCheck
                if val is not True:
                    raise Exception(
                        "illegal extension parameter value '%s' for parameter '%s' of extension '%s'"
                        % (val, p, cls.EXTENSION_NAME)
                    )
                else:
                    server_no_context_takeover = True

            else:
                raise Exception(
                    "illegal extension parameter '%s' for extension '%s'"
                    % (p, cls.EXTENSION_NAME)
                )

        response = cls(client_no_context_takeover, server_no_context_takeover)
        return response

    def __init__(self, client_no_context_takeover, server_no_context_takeover):
        self.client_no_context_takeover = client_no_context_takeover
        self.server_no_context_takeover = server_no_context_takeover

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: JSON serializable representation.
        :rtype: dict
        """
        return {
            "extension": self.EXTENSION_NAME,
            "client_no_context_takeover": self.client_no_context_takeover,
            "server_no_context_takeover": self.server_no_context_takeover,
        }

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: Python string representation.
        :rtype: str
        """
        return (
            "PerMessageBrotliResponse(client_no_context_takeover = %s, server_no_context_takeover = %s)"
            % (self.client_no_context_takeover, self.server_no_context_takeover)
        )


class PerMessageBrotliResponseAccept(
    PerMessageCompressResponseAccept, PerMessageBrotliMixin
):
    """
    Set of parameters with which to accept an `permessage-brotli` response
    from a server by a client.
    """

    def __init__(self, response, no_context_takeover=None):
        """

        :param response: The response being accepted.
        :type response: Instance of :class:`autobahn.compress.PerMessageBrotliResponse`.
        :param no_context_takeover: Override client ("client-to-server direction") context takeover (this must be compatible with response).
        :type no_context_takeover: bool
        """
        if not isinstance(response, PerMessageBrotliResponse):
            raise Exception("invalid type %s for response" % type(response))

        self.response = response

        if no_context_takeover is not None:
            if type(no_context_takeover) != bool:
                raise Exception(
                    "invalid type %s for no_context_takeover"
                    % type(no_context_takeover)
                )

            if response.client_no_context_takeover and not no_context_takeover:
                raise Exception(
                    "invalid value %s for no_context_takeover - server requested feature"
                    % no_context_takeover
                )

        self.no_context_takeover = no_context_takeover

    def __json__(self):
        """
        Returns a JSON serializable object representation.

        :returns: JSON serializable representation.
        :rtype: dict
        """
        return {
            "extension": self.EXTENSION_NAME,
            "response": self.response.__json__(),
            "no_context_takeover": self.no_context_takeover,
        }

    def __repr__(self):
        """
        Returns Python object representation that can be eval'ed to reconstruct the object.

        :returns: Python string representation.
        :rtype: str
        """
        return (
            "PerMessageBrotliResponseAccept(response = %s, no_context_takeover = %s)"
            % (self.response.__repr__(), self.no_context_takeover)
        )


class PerMessageBrotli(PerMessageCompress, PerMessageBrotliMixin):
    """
    `permessage-brotli` WebSocket extension processor.
    """

    @classmethod
    def create_from_response_accept(cls, is_server, accept):
        pmce = cls(
            is_server,
            accept.response.server_no_context_takeover,
            (
                accept.no_context_takeover
                if accept.no_context_takeover is not None
                else accept.response.client_no_context_takeover
            ),
        )
        return pmce

    @classmethod
    def create_from_offer_accept(cls, is_server, accept):
        pmce = cls(
            is_server,
            (
                accept.no_context_takeover
                if accept.no_context_takeover is not None
                else accept.offer.request_no_context_takeover
            ),
            accept.request_no_context_takeover,
        )
        return pmce

    def __init__(
        self, is_server, server_no_context_takeover, client_no_context_takeover
    ):
        self._is_server = is_server
        self.server_no_context_takeover = server_no_context_takeover
        self.client_no_context_takeover = client_no_context_takeover

        self._compressor = None
        self._decompressor = None

    def __json__(self):
        return {
            "extension": self.EXTENSION_NAME,
            "server_no_context_takeover": self.server_no_context_takeover,
            "client_no_context_takeover": self.client_no_context_takeover,
        }

    def __repr__(self):
        return (
            "PerMessageBrotli(is_server = %s, server_no_context_takeover = %s, client_no_context_takeover = %s)"
            % (
                self._is_server,
                self.server_no_context_takeover,
                self.client_no_context_takeover,
            )
        )

    def start_compress_message(self):
        if self._is_server:
            if self._compressor is None or self.server_no_context_takeover:
                self._compressor = brotli.Compressor()
        else:
            if self._compressor is None or self.client_no_context_takeover:
                self._compressor = brotli.Compressor()

    def compress_message_data(self, data):
        return self._compressor.process(data)

    def end_compress_message(self):
        return self._compressor.finish()

    def start_decompress_message(self):
        if self._is_server:
            if self._decompressor is None or self.client_no_context_takeover:
                self._decompressor = brotli.Decompressor()
        else:
            if self._decompressor is None or self.server_no_context_takeover:
                self._decompressor = brotli.Decompressor()

    def decompress_message_data(self, data):
        return self._decompressor.process(data)

    def end_decompress_message(self):
        pass



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

from zope.interface import implementer

from twisted.internet.interfaces import (
    IStreamClientEndpoint,
    IStreamServerEndpoint,
    IStreamServerEndpointStringParser,
)
from twisted.plugin import IPlugin

try:
    from twisted.internet.interfaces import IStreamClientEndpointStringParserWithReactor

    _HAS_REACTOR_ARG = True
except ImportError:
    _HAS_REACTOR_ARG = False
    from twisted.internet.interfaces import (
        IStreamClientEndpointStringParser as IStreamClientEndpointStringParserWithReactor,
    )

from autobahn.twisted.websocket import (
    WrappingWebSocketClientFactory,
    WrappingWebSocketServerFactory,
)
from twisted.internet.endpoints import clientFromString, serverFromString


def _parseOptions(options):
    opts = {}

    if "url" not in options:
        raise Exception("URL needed")
    else:
        opts["url"] = options["url"]

    if "compression" in options:
        value = options["compression"].lower().strip()
        if value == "true":
            opts["enableCompression"] = True
        elif value == "false":
            opts["enableCompression"] = False
        else:
            raise Exception("invalid value '{0}' for compression".format(value))

    if "autofrag" in options:
        try:
            value = int(options["autofrag"])
        except:
            raise Exception(
                "invalid value '{0}' for autofrag".format(options["autofrag"])
            )

        if value < 0:
            raise Exception("negative value '{0}' for autofrag".format(value))

        opts["autoFragmentSize"] = value

    if "subprotocol" in options:
        value = options["subprotocol"].lower().strip()
        opts["subprotocol"] = value

    if "debug" in options:
        value = options["debug"].lower().strip()
        if value == "true":
            opts["debug"] = True
        elif value == "false":
            opts["debug"] = False
        else:
            raise Exception("invalid value '{0}' for debug".format(value))

    return opts


@implementer(IPlugin)
@implementer(IStreamServerEndpointStringParser)
class AutobahnServerParser(object):
    prefix = "autobahn"

    def parseStreamServer(self, reactor, description, **options):
        # The present endpoint plugin is intended to be used as in the
        # following for running a streaming protocol over WebSocket over
        # an underlying stream transport.
        #
        # endpoint = serverFromString(reactor,
        # "autobahn:tcp\:9000\:interface\=0.0.0.0:url=ws\://localhost\:9000:compress=false"
        #
        # This will result in `parseStreamServer` to be called will
        #
        # description == tcp:9000:interface=0.0.0.0
        #
        # and
        #
        # options == {'url': 'ws://localhost:9000', 'compress': 'false'}
        #
        # Essentially, we are using the `\:` escape to coerce the endpoint descriptor
        # of the underlying stream transport into one (first) positional argument.
        #
        # Note that the `\:` within "url" is another form of escaping!
        #
        opts = _parseOptions(options)
        endpoint = serverFromString(reactor, description)
        return AutobahnServerEndpoint(reactor, endpoint, opts)


@implementer(IPlugin)
@implementer(IStreamServerEndpoint)
class AutobahnServerEndpoint(object):
    def __init__(self, reactor, endpoint, options):
        self._reactor = reactor
        self._endpoint = endpoint
        self._options = options

    def listen(self, protocolFactory):
        return self._endpoint.listen(
            WrappingWebSocketServerFactory(
                protocolFactory, reactor=self._reactor, **self._options
            )
        )


# note that for older Twisted before the WithReactor variant, we
# import it under that name so we can share (most of) this
# implementation...
@implementer(IPlugin)
@implementer(IStreamClientEndpointStringParserWithReactor)
class AutobahnClientParser(object):
    prefix = "autobahn"

    def parseStreamClient(self, *args, **options):
        if _HAS_REACTOR_ARG:
            reactor = args[0]
            if len(args) != 2:
                raise RuntimeError(
                    "autobahn: client plugin takes exactly one positional argument"
                )
            description = args[1]
        else:
            from twisted.internet import reactor

            if len(args) != 1:
                raise RuntimeError(
                    "autobahn: client plugin takes exactly one positional argument"
                )
            description = args[0]
        opts = _parseOptions(options)
        endpoint = clientFromString(reactor, description)
        return AutobahnClientEndpoint(reactor, endpoint, opts)


@implementer(IPlugin)
@implementer(IStreamClientEndpoint)
class AutobahnClientEndpoint(object):
    def __init__(self, reactor, endpoint, options):
        self._reactor = reactor
        self._endpoint = endpoint
        self._options = options

    def connect(self, protocolFactory):
        return self._endpoint.connect(
            WrappingWebSocketClientFactory(
                protocolFactory, reactor=self._reactor, **self._options
            )
        )


autobahnServerParser = AutobahnServerParser()
autobahnClientParser = AutobahnClientParser()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class AuthCraRequest(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = AuthCraRequest()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsAuthCraRequest(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # AuthCraRequest
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)


def AuthCraRequestStart(builder):
    builder.StartObject(0)


def Start(builder):
    AuthCraRequestStart(builder)


def AuthCraRequestEnd(builder):
    return builder.EndObject()


def End(builder):
    return AuthCraRequestEnd(builder)



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import os

from cffi import FFI

# Try normal import first (works when package is installed or in editable mode)
try:
    from autobahn.nvx._compile_args import get_compile_args
except ImportError:
    # Fallback for CFFI build time (before package installation)
    # CFFI's setuptools integration runs builder modules via execfile() outside
    # of package context, so normal imports fail. Use importlib to dynamically
    # load the module from file path as a workaround.
    import importlib.util
    import sys

    _path = os.path.join(os.path.dirname(__file__), "_compile_args.py")
    spec = importlib.util.spec_from_file_location("autobahn.nvx._compile_args", _path)
    mod = importlib.util.module_from_spec(spec)
    sys.modules[spec.name] = mod
    spec.loader.exec_module(mod)
    get_compile_args = mod.get_compile_args

ffi = FFI()

ffi.cdef(
    """
    void* nvx_utf8vld_new ();

    void nvx_utf8vld_reset (void* utf8vld);

    int nvx_utf8vld_validate (void* utf8vld, const uint8_t* data, size_t length);

    void nvx_utf8vld_free (void* utf8vld);

    int nvx_utf8vld_set_impl(void* utf8vld, int impl);

    int nvx_utf8vld_get_impl(void* utf8vld);

    size_t nvx_utf8vld_get_current_index (void* utf8vld);

    size_t nvx_utf8vld_get_total_index (void* utf8vld);
"""
)

if "AUTOBAHN_USE_NVX" in os.environ and os.environ["AUTOBAHN_USE_NVX"] in ["1", "true"]:
    optional = False  # :noindex:
else:
    optional = True  # :noindex:

# Get appropriate compiler flags for this build context
# See autobahn.nvx._compile_args for details on architecture baseline selection
# and build context detection (wheel distribution vs. local source install)
extra_compile_args = get_compile_args()

with open(os.path.join(os.path.dirname(__file__), "_utf8validator.c")) as fd:
    c_source = fd.read()
    ffi.set_source(
        "_nvx_utf8validator",
        c_source,
        libraries=[],
        extra_compile_args=extra_compile_args,
        optional=optional,
    )


class Utf8Validator:
    """
    :noindex:
    """

    def __init__(self):
        self.ffi = ffi

        from _nvx_utf8validator import lib

        self.lib = lib

        self._vld = self.ffi.gc(self.lib.nvx_utf8vld_new(), self.lib.nvx_utf8vld_free)
        # print(self.lib.nvx_utf8vld_get_impl(self._vld))

    def reset(self):
        self.lib.nvx_utf8vld_reset(self._vld)

    def validate(self, ba):
        res = self.lib.nvx_utf8vld_validate(self._vld, ba, len(ba))
        current_index = self.lib.nvx_utf8vld_get_current_index(self._vld)
        total_index = self.lib.nvx_utf8vld_get_total_index(self._vld)
        return (res >= 0, res == 0, current_index, total_index)


if __name__ == "__main__":
    ffi.compile()



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

# WAMP Payload Passthru (PPT) application payload serializer (renamed from `Serializer`).
class PPTSerializer(object):
    # Use same serializer (dynamically or statically typed) for the application payload as used on the transport for the WAMP message envelope.
    TRANSPORT = 0
    # Use JSON serializer for dynamically typed application payload.
    JSON = 1
    # Use MsgPack serializer for dynamically typed application payload.
    MSGPACK = 2
    # Use CBOR serializer for dynamically typed application payload.
    CBOR = 3
    # Use UBJSON serializer for dynamically typed application payload.
    UBJSON = 4
    # Raw pass-through of application payload, uninterpreted in any way.
    OPAQUE = 5
    # Use FlatBuffers serializer for statically typed application payload (https://google.github.io/flatbuffers/index.html).
    FLATBUFFERS = 6
    # Use FlexBuffers serializer dynamically typed application payload (https://google.github.io/flatbuffers/flexbuffers.html).
    FLEXBUFFERS = 7



#!/usr/bin/env python3
"""
Modern setup.py that works with pyproject.toml for CFFI module building.
This handles the conditional CFFI module compilation based on environment variables.
"""

import os
from setuptools import setup

# cffi based extension modules to build, currently only NVX
cffi_modules = []
if "AUTOBAHN_USE_NVX" not in os.environ or os.environ["AUTOBAHN_USE_NVX"] not in [
    "0",
    "false",
]:
    cffi_modules.append("autobahn/nvx/_utf8validator.py:ffi")
    cffi_modules.append("autobahn/nvx/_xormasker.py:ffi")

# Include package data from MANIFEST.in
include_package_data = True

# Use setuptools with pyproject.toml for most configuration
setup(
    cffi_modules=cffi_modules,
    include_package_data=include_package_data,
)



import asyncio
import os

# because py.test tries to collect it as a test-case
from unittest.mock import Mock

import pytest
import txaio

from autobahn.asyncio.websocket import WebSocketServerFactory


async def echo_async(what, when):
    await asyncio.sleep(when)
    return what


@pytest.mark.skipif(
    not os.environ.get("USE_ASYNCIO", False), reason="test runs on asyncio only"
)
@pytest.mark.asyncio
async def test_echo_async():
    assert "Hello!" == await echo_async("Hello!", 0)


@pytest.mark.skipif(
    not os.environ.get("USE_ASYNCIO", False), reason="test runs on asyncio only"
)
def test_websocket_custom_loop():
    factory = WebSocketServerFactory(loop=asyncio.new_event_loop())
    server = factory()
    transport = Mock()
    server.connection_made(transport)


@pytest.mark.skipif(
    not os.environ.get("USE_ASYNCIO", False), reason="test runs on asyncio only"
)
def test_async_on_connect_server():
    num = 42
    done = txaio.create_future()
    values = []

    async def foo(x):
        await asyncio.sleep(1)
        return x * x

    async def on_connect(req):
        v = await foo(num)
        values.append(v)
        txaio.resolve(done, req)

    factory = WebSocketServerFactory()
    server = factory()
    server.onConnect = on_connect
    transport = Mock()

    server.connection_made(transport)
    server.data = b"\r\n".join(
        [
            b"GET /ws HTTP/1.1",
            b"Host: www.example.com",
            b"Sec-WebSocket-Version: 13",
            b"Origin: http://www.example.com.malicious.com",
            b"Sec-WebSocket-Extensions: permessage-deflate",
            b"Sec-WebSocket-Key: tXAxWFUqnhi86Ajj7dRY5g==",
            b"Connection: keep-alive, Upgrade",
            b"Upgrade: websocket",
            b"\r\n",  # last string doesn't get a \r\n from join()
        ]
    )
    server.processHandshake()

    asyncio.get_event_loop().run_until_complete(done)

    assert len(values) == 1
    assert values[0] == num * num



###############################################################################
#
# The MIT License (MIT)
#
# Copyright (c) typedef int GmbH
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
###############################################################################

import decimal
import math
import os
import platform
import re
import struct
from binascii import a2b_hex, b2a_hex
from typing import List, Optional, Tuple

from txaio import time_ns

from autobahn.wamp import message
from autobahn.wamp.exception import ProtocolError
from autobahn.wamp.interfaces import IMessage, IObjectSerializer, ISerializer

# note: __all__ must be a list here, since we dynamically
# extend it depending on availability of more serializers
__all__ = ["JsonObjectSerializer", "JsonSerializer", "Serializer"]


SERID_TO_OBJSER = {}
SERID_TO_SER = {}


class Serializer(object):
    """
    Base class for WAMP serializers. A WAMP serializer is the core glue between
    parsed WAMP message objects and the bytes on wire (the transport).
    """

    RATED_MESSAGE_SIZE = 512
    """
    Serialized WAMP message payload size per rated WAMP message.
    """

    # WAMP defines the following 24 message types
    MESSAGE_TYPE_MAP = {
        message.Hello.MESSAGE_TYPE: message.Hello,
        message.Welcome.MESSAGE_TYPE: message.Welcome,
        message.Abort.MESSAGE_TYPE: message.Abort,
        message.Challenge.MESSAGE_TYPE: message.Challenge,
        message.Authenticate.MESSAGE_TYPE: message.Authenticate,
        message.Goodbye.MESSAGE_TYPE: message.Goodbye,
        message.Error.MESSAGE_TYPE: message.Error,
        message.Publish.MESSAGE_TYPE: message.Publish,
        message.Published.MESSAGE_TYPE: message.Published,
        message.Subscribe.MESSAGE_TYPE: message.Subscribe,
        message.Subscribed.MESSAGE_TYPE: message.Subscribed,
        message.Unsubscribe.MESSAGE_TYPE: message.Unsubscribe,
        message.Unsubscribed.MESSAGE_TYPE: message.Unsubscribed,
        message.Event.MESSAGE_TYPE: message.Event,
        message.EventReceived.MESSAGE_TYPE: message.EventReceived,
        message.Call.MESSAGE_TYPE: message.Call,
        message.Cancel.MESSAGE_TYPE: message.Cancel,
        message.Result.MESSAGE_TYPE: message.Result,
        message.Register.MESSAGE_TYPE: message.Register,
        message.Registered.MESSAGE_TYPE: message.Registered,
        message.Unregister.MESSAGE_TYPE: message.Unregister,
        message.Unregistered.MESSAGE_TYPE: message.Unregistered,
        message.Invocation.MESSAGE_TYPE: message.Invocation,
        message.Interrupt.MESSAGE_TYPE: message.Interrupt,
        message.Yield.MESSAGE_TYPE: message.Yield,
    }
    """
    Mapping of WAMP message type codes to WAMP message classes.
    """

    def __init__(self, serializer):
        """

        :param serializer: The object serializer to use for WAMP wire-level serialization.
        :type serializer: An object that implements :class:`autobahn.interfaces.IObjectSerializer`.
        """
        self._serializer = serializer
        # Store back-reference so Message.build() can access parent ISerializer
        self._serializer._parent_serializer = self

        self._stats_reset = time_ns()
        self._stats_cycle = 0

        self._serialized_bytes = 0
        self._serialized_messages = 0
        self._serialized_rated_messages = 0

        self._unserialized_bytes = 0
        self._unserialized_messages = 0
        self._unserialized_rated_messages = 0

        self._autoreset_rated_messages = None
        self._autoreset_duration = None
        self._autoreset_callback = None

    def stats_reset(self):
        """
        Get serializer statistics: timestamp when statistics were last reset.

        :return: Last reset time of statistics (UTC, ns since Unix epoch)
        :rtype: int
        """
        return self._stats_reset

    def stats_bytes(self):
        """
        Get serializer statistics: bytes (serialized + unserialized).

        :return: Number of bytes.
        :rtype: int
        """
        return self._serialized_bytes + self._unserialized_bytes

    def stats_messages(self):
        """
        Get serializer statistics: messages (serialized + unserialized).

        :return: Number of messages.
        :rtype: int
        """
        return self._serialized_messages + self._unserialized_messages

    def stats_rated_messages(self):
        """
        Get serializer statistics: rated messages (serialized + unserialized).

        :return: Number of rated messages.
        :rtype: int
        """
        return self._serialized_rated_messages + self._unserialized_rated_messages

    def set_stats_autoreset(self, rated_messages, duration, callback, reset_now=False):
        """
        Configure a user callback invoked when accumulated stats hit specified threshold.
        When the specified number of rated messages have been processed or the specified duration
        has passed, statistics are automatically reset, and the last statistics is provided to
        the user callback.

        :param rated_messages: Number of rated messages that should trigger an auto-reset.
        :type rated_messages: int

        :param duration: Duration in ns that when passed will trigger an auto-reset.
        :type duration: int

        :param callback: User callback to be invoked when statistics are auto-reset. The function
            will be invoked with a single positional argument: the accumulated statistics before the reset.
        :type callback: callable
        """
        assert rated_messages is None or type(rated_messages) == int
        assert duration is None or type(duration) == int
        assert rated_messages or duration
        assert callable(callback)

        self._autoreset_rated_messages = rated_messages
        self._autoreset_duration = duration
        self._autoreset_callback = callback

        # maybe auto-reset and trigger user callback ..
        if self._autoreset_callback and reset_now:
            stats = self.stats(reset=True)
            self._autoreset_callback(stats)
            return stats

    def stats(self, reset=True, details=False):
        """
        Get (and reset) serializer statistics.

        :param reset: If ``True``, reset the serializer statistics.
        :type reset: bool

        :param details: If ``True``, return detailed statistics split up by serialization/unserialization.
        :type details: bool

        :return: Serializer statistics, eg:

            .. code-block:: json

                {
                    "timestamp": 1574156576688704693,
                    "duration": 34000000000,
                    "bytes": 0,
                    "messages": 0,
                    "rated_messages": 0
                }

        :rtype: dict
        """
        assert type(reset) == bool
        assert type(details) == bool

        self._stats_cycle += 1

        if details:
            data = {
                "cycle": self._stats_cycle,
                "serializer": self.SERIALIZER_ID,
                "timestamp": self._stats_reset,
                "duration": time_ns() - self._stats_reset,
                "serialized": {
                    "bytes": self._serialized_bytes,
                    "messages": self._serialized_messages,
                    "rated_messages": self._serialized_rated_messages,
                },
                "unserialized": {
                    "bytes": self._unserialized_bytes,
                    "messages": self._unserialized_messages,
                    "rated_messages": self._unserialized_rated_messages,
                },
            }
        else:
            data = {
                "cycle": self._stats_cycle,
                "serializer": self.SERIALIZER_ID,
                "timestamp": self._stats_reset,
                "duration": time_ns() - self._stats_reset,
                "bytes": self._serialized_bytes + self._unserialized_bytes,
                "messages": self._serialized_messages + self._unserialized_messages,
                "rated_messages": self._serialized_rated_messages
                + self._unserialized_rated_messages,
            }
        if reset:
            self._serialized_bytes = 0
            self._serialized_messages = 0
            self._serialized_rated_messages = 0
            self._unserialized_bytes = 0
            self._unserialized_messages = 0
            self._unserialized_rated_messages = 0
            self._stats_reset = time_ns()
        return data

    def serialize(self, msg: IMessage) -> Tuple[bytes, bool]:
        """
        Implements :func:`autobahn.wamp.interfaces.ISerializer.serialize`
        """
        data, is_binary = msg.serialize(self._serializer), self._serializer.BINARY

        # maintain statistics for serialized WAMP message data
        self._serialized_bytes += len(data)
        self._serialized_messages += 1
        self._serialized_rated_messages += int(
            math.ceil(float(len(data)) / self.RATED_MESSAGE_SIZE)
        )

        # maybe auto-reset and trigger user callback ..
        if self._autoreset_callback and (
            (
                self._autoreset_duration
                and (time_ns() - self._stats_reset) >= self._autoreset_duration
            )
            or (
                self._autoreset_rated_messages
                and self.stats_rated_messages() >= self._autoreset_rated_messages
            )
        ):
            stats = self.stats(reset=True)
            self._autoreset_callback(stats)

        return data, is_binary

    def serialize_payload(self, data):
        """
        Serialize application payload data (args/kwargs/payload).

        Uses the payload serializer configured for this transport serializer.
        For traditional serializers (JSON, CBOR, MsgPack, UBJSON), this is the
        same as the envelope serializer. For FlatBuffers, this can be different.

        :param data: The data to serialize (list for args, dict for kwargs).
        :return: Serialized bytes.
        :rtype: bytes
        """
        # FlatBuffersSerializer has _payload_serializer (separate from envelope)
        # Traditional serializers use _serializer (same for envelope and payload)
        payload_ser = getattr(self, "_payload_serializer", self._serializer)
        return payload_ser.serialize(data)

    def unserialize(
        self, payload: bytes, isBinary: Optional[bool] = None
    ) -> List[IMessage]:
        """
        Implements :func:`autobahn.wamp.interfaces.ISerializer.unserialize`
        """
        if isBinary is not None:
            if isBinary != self._serializer.BINARY:
                raise ProtocolError(
                    "invalid serialization of WAMP message (binary {0}, but expected {1})".format(
                        isBinary, self._serializer.BINARY
                    )
                )
        try:
            raw_msgs = self._serializer.unserialize(payload)
        except Exception as e:
            raise ProtocolError(
                "invalid serialization of WAMP message: {0} {1}".format(
                    type(e).__name__, e
                )
            )

        if self._serializer.NAME == "flatbuffers":
            msgs = raw_msgs
        else:
            msgs = []
            for raw_msg in raw_msgs:
                if type(raw_msg) != list:
                    raise ProtocolError(
                        "invalid type {0} for WAMP message".format(type(raw_msg))
                    )

                if len(raw_msg) == 0:
                    raise ProtocolError("missing message type in WAMP message")

                message_type = raw_msg[0]

                if type(message_type) != int:
                    raise ProtocolError(
                        "invalid type {0} for WAMP message type".format(
                            type(message_type)
                        )
                    )

                Klass = self.MESSAGE_TYPE_MAP.get(message_type)

                if Klass is None:
                    raise ProtocolError(
                        "invalid WAMP message type {0}".format(message_type)
                    )

                # this might again raise `ProtocolError` ..
                msg = Klass.parse(raw_msg)

                msgs.append(msg)

        # maintain statistics for unserialized WAMP message data
        self._unserialized_bytes += len(payload)
        self._unserialized_messages += len(msgs)
        self._unserialized_rated_messages += int(
            math.ceil(float(len(payload)) / self.RATED_MESSAGE_SIZE)
        )

        # maybe auto-reset and trigger user callback ..
        if self._autoreset_callback and (
            (
                self._autoreset_duration
                and (time_ns() - self._stats_reset) >= self._autoreset_duration
            )
            or (
                self._autoreset_rated_messages
                and self.stats_rated_messages() >= self._autoreset_rated_messages
            )
        ):
            stats = self.stats(reset=True)
            self._autoreset_callback(stats)

        return msgs


# JSON serialization is always supported
_USE_UJSON = "AUTOBAHN_USE_UJSON" in os.environ
if _USE_UJSON:
    try:
        import ujson

        _USE_UJSON = True
    except ImportError:
        import json

        _USE_UJSON = False
else:
    import json


if _USE_UJSON:
    # ujson doesn't support plugging into the JSON string parsing machinery ..
    print(
        "WARNING: Autobahn is using ujson accelerated JSON module - will run faster,\nbut only on CPython and will loose ability to transport binary payload transparently!"
    )
    _loads = ujson.loads
    _dumps = ujson.dumps
    _json = ujson
else:
    # print('Notice: Autobahn is using json built-in standard library module for JSON serialization')
    import base64

    class _WAMPJsonEncoder(json.JSONEncoder):
        def __init__(self, *args, **kwargs):
            if "use_binary_hex_encoding" in kwargs:
                self._use_binary_hex_encoding = kwargs["use_binary_hex_encoding"]
                del kwargs["use_binary_hex_encoding"]
            else:
                self._use_binary_hex_encoding = False
            json.JSONEncoder.__init__(self, *args, **kwargs)

        def default(self, obj):
            if isinstance(obj, bytes):
                if self._use_binary_hex_encoding:
                    return "0x" + b2a_hex(obj).decode("ascii")
                else:
                    return "\x00" + base64.b64encode(obj).decode("ascii")
            elif isinstance(obj, decimal.Decimal):
                return str(obj)
            else:
                return json.JSONEncoder.default(self, obj)

    #
    # the following is a hack. see http://bugs.python.org/issue29992
    #

    from json import scanner
    from json.decoder import scanstring

    _DEC_MATCH = re.compile(r"^[\+\-E\.0-9]+$")

    class _WAMPJsonDecoder(json.JSONDecoder):
        def __init__(self, *args, **kwargs):
            if "use_binary_hex_encoding" in kwargs:
                self._use_binary_hex_encoding = kwargs["use_binary_hex_encoding"]
                del kwargs["use_binary_hex_encoding"]
            else:
                self._use_binary_hex_encoding = False

            if "use_decimal_from_str" in kwargs:
                self._use_decimal_from_str = kwargs["use_decimal_from_str"]
                del kwargs["use_decimal_from_str"]
            else:
                self._use_decimal_from_str = False

            if "use_decimal_from_float" in kwargs:
                self._use_decimal_from_float = kwargs["use_decimal_from_float"]
                del kwargs["use_decimal_from_float"]
                if self._use_decimal_from_float:
                    kwargs["parse_float"] = decimal.Decimal
            else:
                self._use_decimal_from_str = False

            json.JSONDecoder.__init__(self, *args, **kwargs)

            def _parse_string(*args, **kwargs):
                s, idx = scanstring(*args, **kwargs)
                if self._use_binary_hex_encoding:
                    if s and s[0:2] == "0x":
                        s = a2b_hex(s[2:])
                        return s, idx
                else:
                    if s and s[0] == "\x00":
                        s = base64.b64decode(s[1:])
                        return s, idx
                if self._use_decimal_from_str and _DEC_MATCH.match(s):
                    try:
                        s = decimal.Decimal(s)
                        return s, idx
                    except decimal.InvalidOperation:
                        pass
                return s, idx

            self.parse_string = _parse_string

            # we need to recreate the internal scan function ..
            self.scan_once = scanner.py_make_scanner(self)

            # .. and we have to explicitly use the Py version,
            # not the C version, as the latter won't work
            # self.scan_once = scanner.make_scanner(self)

    def _loads(
        s,
        use_binary_hex_encoding=False,
        use_decimal_from_str=False,
        use_decimal_from_float=False,
    ):
        return json.loads(
            s,
            use_binary_hex_encoding=use_binary_hex_encoding,
            use_decimal_from_str=use_decimal_from_str,
            use_decimal_from_float=use_decimal_from_float,
            cls=_WAMPJsonDecoder,
        )

    def _dumps(obj, use_binary_hex_encoding=False):
        return json.dumps(
            obj,
            separators=(",", ":"),
            ensure_ascii=False,
            sort_keys=False,
            use_binary_hex_encoding=use_binary_hex_encoding,
            cls=_WAMPJsonEncoder,
        )

    _json = json


class JsonObjectSerializer(object):
    JSON_MODULE = _json
    """
    The JSON module used (now only stdlib).
    """

    NAME = "json"

    BINARY = False

    def __init__(
        self,
        batched=False,
        use_binary_hex_encoding=False,
        use_decimal_from_str=False,
        use_decimal_from_float=False,
    ):
        """

        :param batched: Flag that controls whether serializer operates in batched mode.
        :type batched: bool

        :param use_binary_hex_encoding: Flag to enable HEX encoding prefixed with ``"0x"``,
            otherwise prefix binaries with a ``\0`` byte.
        :type use_binary_hex_encoding: bool

        :param use_decimal_from_str: Flag to automatically encode Decimals as strings, and
            to try to parse strings as Decimals.
        :type use_decimal_from_str: bool
        """
        self._batched = batched
        self._use_binary_hex_encoding = use_binary_hex_encoding
        self._use_decimal_from_str = use_decimal_from_str
        self._use_decimal_from_float = use_decimal_from_float

    def serialize(self, obj):
        """
        Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.serialize`
        """
        s = _dumps(obj, use_binary_hex_encoding=self._use_binary_hex_encoding)
        if isinstance(s, str):
            s = s.encode("utf8")
        if self._batched:
            return s + b"\30"
        else:
            return s

    def unserialize(self, payload):
        """
        Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.unserialize`
        """
        if self._batched:
            chunks = payload.split(b"\30")[:-1]
        else:
            chunks = [payload]
        if len(chunks) == 0:
            raise Exception("batch format error")
        return [
            _loads(
                data.decode("utf8"),
                use_binary_hex_encoding=self._use_binary_hex_encoding,
                use_decimal_from_str=self._use_decimal_from_str,
                use_decimal_from_float=self._use_decimal_from_float,
            )
            for data in chunks
        ]


IObjectSerializer.register(JsonObjectSerializer)
SERID_TO_OBJSER[JsonObjectSerializer.NAME] = JsonObjectSerializer


class JsonSerializer(Serializer):
    SERIALIZER_ID = "json"
    """
    ID used as part of the WebSocket subprotocol name to identify the
    serializer with WAMP-over-WebSocket.
    """

    RAWSOCKET_SERIALIZER_ID = 1
    """
    ID used in lower four bits of second octet in RawSocket opening
    handshake identify the serializer with WAMP-over-RawSocket.
    """

    PAYLOAD_SERIALIZER_ID = "json"
    """
    Serializer for application payload. For JSON transport, both envelope
    and payload use JSON serialization.
    """

    MIME_TYPE = "application/json"
    """
    MIME type announced in HTTP request/response headers when running
    WAMP-over-Longpoll HTTP fallback.
    """

    def __init__(
        self, batched=False, use_binary_hex_encoding=False, use_decimal_from_str=False
    ):
        """
        Ctor.

        :param batched: Flag to control whether to put this serialized into batched mode.
        :type batched: bool
        """
        Serializer.__init__(
            self,
            JsonObjectSerializer(
                batched=batched,
                use_binary_hex_encoding=use_binary_hex_encoding,
                use_decimal_from_str=use_decimal_from_str,
            ),
        )
        if batched:
            self.SERIALIZER_ID = "json.batched"


ISerializer.register(JsonSerializer)
SERID_TO_SER[JsonSerializer.SERIALIZER_ID] = JsonSerializer


_HAS_MSGPACK = False
_USE_UMSGPACK = (
    platform.python_implementation() == "PyPy" or "AUTOBAHN_USE_UMSGPACK" in os.environ
)

if not _USE_UMSGPACK:
    try:
        # on CPython, use an impl. with native extension:
        # https://pypi.org/project/msgpack/
        # https://github.com/msgpack/msgpack-python
        import msgpack
    except ImportError:
        pass
    else:
        _HAS_MSGPACK = True
        _packb = lambda obj: msgpack.packb(obj, use_bin_type=True)  # noqa
        _unpackb = lambda data: msgpack.unpackb(data, raw=False)  # noqa
        _msgpack = msgpack
        # print('Notice: Autobahn is using msgpack library (with native extension, best on CPython) for MessagePack serialization')
else:
    try:
        # on PyPy in particular, use a pure python impl.:
        # https://pypi.python.org/pypi/u-msgpack-python
        # https://github.com/vsergeev/u-msgpack-python
        import umsgpack
    except ImportError:
        pass
    else:
        _HAS_MSGPACK = True
        _packb = umsgpack.packb
        _unpackb = umsgpack.unpackb
        _msgpack = umsgpack
        # print('Notice: Autobahn is using umsgpack library (pure Python, best on PyPy) for MessagePack serialization')


if _HAS_MSGPACK:

    class MsgPackObjectSerializer(object):
        NAME = "msgpack"

        MSGPACK_MODULE = _msgpack

        BINARY = True
        """
        Flag that indicates whether this serializer needs a binary clean transport.
        """

        def __init__(self, batched=False):
            """

            :param batched: Flag that controls whether serializer operates in batched mode.
            :type batched: bool
            """
            self._batched = batched

        def serialize(self, obj):
            """
            Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.serialize`
            """
            data = _packb(obj)
            if self._batched:
                return struct.pack("!L", len(data)) + data
            else:
                return data

        def unserialize(self, payload):
            """
            Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.unserialize`
            """

            if self._batched:
                msgs = []
                N = len(payload)
                i = 0
                while i < N:
                    # read message length prefix
                    if i + 4 > N:
                        raise Exception("batch format error [1]")
                    l = struct.unpack("!L", payload[i : i + 4])[0]

                    # read message data
                    if i + 4 + l > N:
                        raise Exception("batch format error [2]")
                    data = payload[i + 4 : i + 4 + l]

                    # append parsed raw message
                    msgs.append(_unpackb(data))

                    # advance until everything consumed
                    i = i + 4 + l

                if i != N:
                    raise Exception("batch format error [3]")
                return msgs

            else:
                unpacked = _unpackb(payload)
                return [unpacked]

    IObjectSerializer.register(MsgPackObjectSerializer)

    __all__.append("MsgPackObjectSerializer")
    SERID_TO_OBJSER[MsgPackObjectSerializer.NAME] = MsgPackObjectSerializer

    class MsgPackSerializer(Serializer):
        SERIALIZER_ID = "msgpack"
        """
        ID used as part of the WebSocket subprotocol name to identify the
        serializer with WAMP-over-WebSocket.
        """

        RAWSOCKET_SERIALIZER_ID = 2
        """
        ID used in lower four bits of second octet in RawSocket opening
        handshake identify the serializer with WAMP-over-RawSocket.
        """

        PAYLOAD_SERIALIZER_ID = "msgpack"
        """
        Serializer for application payload. For MessagePack transport, both envelope
        and payload use MessagePack serialization.
        """

        MIME_TYPE = "application/x-msgpack"
        """
        MIME type announced in HTTP request/response headers when running
        WAMP-over-Longpoll HTTP fallback.
        """

        def __init__(self, batched=False):
            """
            Ctor.

            :param batched: Flag to control whether to put this serialized into batched mode.
            :type batched: bool
            """
            Serializer.__init__(self, MsgPackObjectSerializer(batched=batched))
            if batched:
                self.SERIALIZER_ID = "msgpack.batched"

    ISerializer.register(MsgPackSerializer)
    SERID_TO_SER[MsgPackSerializer.SERIALIZER_ID] = MsgPackSerializer

    __all__.append("MsgPackSerializer")


_HAS_CBOR = False


try:
    import cbor2
except ImportError:
    pass
else:
    _HAS_CBOR = True
    _cbor_loads = cbor2.loads
    _cbor_dumps = cbor2.dumps
    _cbor = cbor2


if _HAS_CBOR:

    class CBORObjectSerializer(object):
        """
        CBOR serializer based on `cbor2 <https://github.com/agronholm/cbor2>`_.

        This CBOR serializer has proper support for arbitrary precision decimals,
        via tagged decimal fraction encoding, as described in
        `RFC7049 section 2.4.3 <https://datatracker.ietf.org/doc/html/rfc7049#section-2.4.3>`_.
        """

        NAME = "cbor"

        CBOR_MODULE = _cbor

        BINARY = True
        """
        Flag that indicates whether this serializer needs a binary clean transport.
        """

        def __init__(self, batched=False):
            """

            :param batched: Flag that controls whether serializer operates in batched mode.
            :type batched: bool
            """
            self._batched = batched

        def serialize(self, obj):
            """
            Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.serialize`
            """
            data = _cbor_dumps(obj)
            if self._batched:
                return struct.pack("!L", len(data)) + data
            else:
                return data

        def unserialize(self, payload):
            """
            Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.unserialize`
            """

            if self._batched:
                msgs = []
                N = len(payload)
                i = 0
                while i < N:
                    # read message length prefix
                    if i + 4 > N:
                        raise Exception("batch format error [1]")
                    l = struct.unpack("!L", payload[i : i + 4])[0]

                    # read message data
                    if i + 4 + l > N:
                        raise Exception("batch format error [2]")
                    data = payload[i + 4 : i + 4 + l]

                    # append parsed raw message
                    msgs.append(_cbor_loads(data))

                    # advance until everything consumed
                    i = i + 4 + l

                if i != N:
                    raise Exception("batch format error [3]")
                return msgs

            else:
                unpacked = _cbor_loads(payload)
                return [unpacked]

    IObjectSerializer.register(CBORObjectSerializer)
    SERID_TO_OBJSER[CBORObjectSerializer.NAME] = CBORObjectSerializer

    __all__.append("CBORObjectSerializer")

    class CBORSerializer(Serializer):
        SERIALIZER_ID = "cbor"
        """
        ID used as part of the WebSocket subprotocol name to identify the
        serializer with WAMP-over-WebSocket.
        """

        RAWSOCKET_SERIALIZER_ID = 3
        """
        ID used in lower four bits of second octet in RawSocket opening
        handshake identify the serializer with WAMP-over-RawSocket.
        """

        PAYLOAD_SERIALIZER_ID = "cbor"
        """
        Serializer for application payload. For CBOR transport, both envelope
        and payload use CBOR serialization.
        """

        MIME_TYPE = "application/cbor"
        """
        MIME type announced in HTTP request/response headers when running
        WAMP-over-Longpoll HTTP fallback.
        """

        def __init__(self, batched=False):
            """
            Ctor.

            :param batched: Flag to control whether to put this serialized into batched mode.
            :type batched: bool
            """
            Serializer.__init__(self, CBORObjectSerializer(batched=batched))
            if batched:
                self.SERIALIZER_ID = "cbor.batched"

    ISerializer.register(CBORSerializer)
    SERID_TO_SER[CBORSerializer.SERIALIZER_ID] = CBORSerializer

    __all__.append("CBORSerializer")


# UBJSON serialization depends on the `py-ubjson` package being available
# https://pypi.python.org/pypi/py-ubjson
# https://github.com/Iotic-Labs/py-ubjson
try:
    import ubjson
except ImportError:
    pass
else:
    # print('Notice: Autobahn is using ubjson module for UBJSON serialization')

    class UBJSONObjectSerializer(object):
        NAME = "ubjson"

        UBJSON_MODULE = ubjson

        BINARY = True
        """
        Flag that indicates whether this serializer needs a binary clean transport.
        """

        def __init__(self, batched=False):
            """

            :param batched: Flag that controls whether serializer operates in batched mode.
            :type batched: bool
            """
            self._batched = batched

        def serialize(self, obj):
            """
            Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.serialize`
            """
            data = ubjson.dumpb(obj)
            if self._batched:
                return struct.pack("!L", len(data)) + data
            else:
                return data

        def unserialize(self, payload):
            """
            Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.unserialize`
            """

            if self._batched:
                msgs = []
                N = len(payload)
                i = 0
                while i < N:
                    # read message length prefix
                    if i + 4 > N:
                        raise Exception("batch format error [1]")
                    l = struct.unpack("!L", payload[i : i + 4])[0]

                    # read message data
                    if i + 4 + l > N:
                        raise Exception("batch format error [2]")
                    data = payload[i + 4 : i + 4 + l]

                    # append parsed raw message
                    msgs.append(ubjson.loadb(data))

                    # advance until everything consumed
                    i = i + 4 + l

                if i != N:
                    raise Exception("batch format error [3]")
                return msgs

            else:
                unpacked = ubjson.loadb(payload)
                return [unpacked]

    IObjectSerializer.register(UBJSONObjectSerializer)
    SERID_TO_OBJSER[UBJSONObjectSerializer.NAME] = UBJSONObjectSerializer

    __all__.append("UBJSONObjectSerializer")

    class UBJSONSerializer(Serializer):
        SERIALIZER_ID = "ubjson"
        """
        ID used as part of the WebSocket subprotocol name to identify the
        serializer with WAMP-over-WebSocket.
        """

        RAWSOCKET_SERIALIZER_ID = 4
        """
        ID used in lower four bits of second octet in RawSocket opening
        handshake identify the serializer with WAMP-over-RawSocket.
        """

        PAYLOAD_SERIALIZER_ID = "ubjson"
        """
        Serializer for application payload. For UBJSON transport, both envelope
        and payload use UBJSON serialization.
        """

        MIME_TYPE = "application/ubjson"
        """
        MIME type announced in HTTP request/response headers when running
        WAMP-over-Longpoll HTTP fallback.
        """

        def __init__(self, batched=False):
            """
            Ctor.

            :param batched: Flag to control whether to put this serialized into batched mode.
            :type batched: bool
            """
            Serializer.__init__(self, UBJSONObjectSerializer(batched=batched))
            if batched:
                self.SERIALIZER_ID = "ubjson.batched"

    ISerializer.register(UBJSONSerializer)
    SERID_TO_SER[UBJSONSerializer.SERIALIZER_ID] = UBJSONSerializer

    __all__.append("UBJSONSerializer")


_HAS_FLATBUFFERS = False
try:
    import flatbuffers  # noqa
    from autobahn.wamp import message_fbs
except ImportError:
    pass
else:
    _HAS_FLATBUFFERS = True


if _HAS_FLATBUFFERS:

    class FlatBuffersObjectSerializer(object):
        NAME = "flatbuffers"

        FLATBUFFERS_MODULE = flatbuffers

        BINARY = True
        """
        Flag that indicates whether this serializer needs a binary clean transport.
        """

        MESSAGE_TYPE_MAP = {
            # Category 4: Both Payload and Forwarding
            message_fbs.MessageType.ERROR: (message_fbs.Error, message.Error),
            message_fbs.MessageType.EVENT: (message_fbs.Event, message.Event),
            message_fbs.MessageType.PUBLISH: (message_fbs.Publish, message.Publish),
            message_fbs.MessageType.CALL: (message_fbs.Call, message.Call),
            message_fbs.MessageType.RESULT: (message_fbs.Result, message.Result),
            message_fbs.MessageType.INVOCATION: (message_fbs.Invocation, message.Invocation),
            message_fbs.MessageType.YIELD: (message_fbs.Yield, message.Yield),

            # Category 1: Session lifecycle messages
            message_fbs.MessageType.HELLO: (message_fbs.HelloGen.Hello, message.Hello),
            message_fbs.MessageType.WELCOME: (message_fbs.WelcomeGen.Welcome, message.Welcome),
            message_fbs.MessageType.ABORT: (message_fbs.AbortGen.Abort, message.Abort),
            message_fbs.MessageType.CHALLENGE: (message_fbs.ChallengeGen.Challenge, message.Challenge),
            message_fbs.MessageType.AUTHENTICATE: (message_fbs.AuthenticateGen.Authenticate, message.Authenticate),
            message_fbs.MessageType.GOODBYE: (message_fbs.GoodbyeGen.Goodbye, message.Goodbye),

            # Category 1: PubSub messages
            message_fbs.MessageType.SUBSCRIBE: (message_fbs.SubscribeGen.Subscribe, message.Subscribe),
            message_fbs.MessageType.SUBSCRIBED: (message_fbs.SubscribedGen.Subscribed, message.Subscribed),
            message_fbs.MessageType.PUBLISHED: (message_fbs.PublishedGen.Published, message.Published),
            message_fbs.MessageType.UNSUBSCRIBE: (message_fbs.UnsubscribeGen.Unsubscribe, message.Unsubscribe),
            message_fbs.MessageType.UNSUBSCRIBED: (message_fbs.UnsubscribedGen.Unsubscribed, message.Unsubscribed),

            # Category 1: RPC messages
            message_fbs.MessageType.REGISTER: (message_fbs.RegisterGen.Register, message.Register),
            message_fbs.MessageType.REGISTERED: (message_fbs.RegisteredGen.Registered, message.Registered),
            message_fbs.MessageType.UNREGISTER: (message_fbs.UnregisterGen.Unregister, message.Unregister),
            message_fbs.MessageType.UNREGISTERED: (message_fbs.UnregisteredGen.Unregistered, message.Unregistered),

            # Category 3: Forwarding Only messages
            message_fbs.MessageType.EVENT_RECEIVED: (message_fbs.EventReceivedGen.EventReceived, message.EventReceived),
            message_fbs.MessageType.CANCEL: (message_fbs.CancelGen.Cancel, message.Cancel),
            message_fbs.MessageType.INTERRUPT: (message_fbs.InterruptGen.Interrupt, message.Interrupt),
        }

        def __init__(self, batched=False):
            """

            :param batched: Flag that controls whether serializer operates in batched mode.
            :type batched: bool
            """
            assert not batched, (
                "WAMP-FlatBuffers serialization does not support message batching currently"
            )
            self._batched = batched

        def serialize(self, obj):
            """
            Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.serialize`
            """
            raise NotImplementedError()

        def unserialize(self, payload):
            """
            Implements :func:`autobahn.wamp.interfaces.IObjectSerializer.unserialize`
            """
            union_msg = message_fbs.Message.Message.GetRootAsMessage(payload, 0)
            msg_type = union_msg.MsgType()

            if msg_type in self.MESSAGE_TYPE_MAP:
                fbs_klass, wamp_klass = self.MESSAGE_TYPE_MAP[msg_type]
                fbs_msg = fbs_klass()
                _tab = union_msg.Msg()
                fbs_msg.Init(_tab.Bytes, _tab.Pos)
                msg = wamp_klass(from_fbs=fbs_msg)
                return [msg]
            else:
                raise NotImplementedError(
                    "message type {} not yet implemented for WAMP-FlatBuffers".format(
                        msg_type
                    )
                )

    IObjectSerializer.register(FlatBuffersObjectSerializer)

    __all__.append("FlatBuffersObjectSerializer")
    SERID_TO_OBJSER[FlatBuffersObjectSerializer.NAME] = FlatBuffersObjectSerializer

    class FlatBuffersSerializer(Serializer):
        SERIALIZER_ID = "flatbuffers"
        """
        ID used as part of the WebSocket subprotocol name to identify the
        serializer with WAMP-over-WebSocket.
        """

        RAWSOCKET_SERIALIZER_ID = 5
        """
        ID used in lower four bits of second octet in RawSocket opening
        handshake identify the serializer with WAMP-over-RawSocket.
        """

        MIME_TYPE = "application/x-flatbuffers"
        """
        MIME type announced in HTTP request/response headers when running
        WAMP-over-Longpoll HTTP fallback.
        """

        def __init__(self, batched=False, payload_serializer="cbor"):
            """

            :param batched: Flag to control whether to put this serialized into batched mode.
            :type batched: bool
            :param payload_serializer: Serializer ID for application payload (args/kwargs/payload).
                Can be "json", "msgpack", "cbor", "ubjson", or "flatbuffers". Defaults to "cbor".
            :type payload_serializer: str
            """
            Serializer.__init__(self, FlatBuffersObjectSerializer(batched=batched))
            if batched:
                self.SERIALIZER_ID = "flatbuffers.batched"

            # Store payload serializer ID and create instance
            self._payload_serializer_id = payload_serializer
            if payload_serializer in SERID_TO_OBJSER:
                payload_ser_class = SERID_TO_OBJSER[payload_serializer]
                self._payload_serializer = payload_ser_class()
            else:
                raise ValueError(
                    f"Unknown payload serializer '{payload_serializer}'. "
                    f"Available: {sorted(SERID_TO_OBJSER.keys())}"
                )

        @property
        def PAYLOAD_SERIALIZER_ID(self):
            """
            Serializer for application payload. For FlatBuffers transport, this can
            differ from the envelope serializer to enable composition patterns like
            FlatBuffers envelope with CBOR payload.
            """
            return self._payload_serializer_id

    ISerializer.register(FlatBuffersSerializer)
    SERID_TO_SER[FlatBuffersSerializer.SERIALIZER_ID] = FlatBuffersSerializer

    __all__.append("FlatBuffersSerializer")


def create_transport_serializer(serializer_id):
    batched = False
    if "." in serializer_id:
        l = serializer_id.split(".")
        serializer_id = l[0]
        if len(l) > 1 and l[1] == "batched":
            batched = True

    if serializer_id in SERID_TO_SER:
        return SERID_TO_SER[serializer_id](batched=batched)
    else:
        raise RuntimeError(
            'could not create serializer for "{}" (available: {})'.format(
                serializer_id, sorted(SERID_TO_SER.keys())
            )
        )


def create_transport_serializers(transport):
    """
    Create a list of serializers to use with a WAMP protocol factory.
    """
    serializers = []
    for serializer_id in transport.serializers:
        serializers.append(create_transport_serializer(serializer_id))
    return serializers



# automatically generated by the FlatBuffers compiler, do not modify

# namespace: proto

import flatbuffers
from flatbuffers.compat import import_numpy

np = import_numpy()


class SubscriberFeatures(object):
    __slots__ = ["_tab"]

    @classmethod
    def GetRootAs(cls, buf, offset=0):
        n = flatbuffers.encode.Get(flatbuffers.packer.uoffset, buf, offset)
        x = SubscriberFeatures()
        x.Init(buf, n + offset)
        return x

    @classmethod
    def GetRootAsSubscriberFeatures(cls, buf, offset=0):
        """This method is deprecated. Please switch to GetRootAs."""
        return cls.GetRootAs(buf, offset)

    # SubscriberFeatures
    def Init(self, buf, pos):
        self._tab = flatbuffers.table.Table(buf, pos)

    # SubscriberFeatures
    def PublisherIdentification(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(4))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # SubscriberFeatures
    def PatternBasedSubscription(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(6))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # SubscriberFeatures
    def PublicationTrustlevels(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(8))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # SubscriberFeatures
    def SubscriptionRevocation(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(10))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # SubscriberFeatures
    def EventHistory(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(12))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # SubscriberFeatures
    def AcknowledgeSubscriberReceived(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(14))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # SubscriberFeatures
    def PayloadTransparency(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(16))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False

    # SubscriberFeatures
    def PayloadEncryptionCryptobox(self):
        o = flatbuffers.number_types.UOffsetTFlags.py_type(self._tab.Offset(18))
        if o != 0:
            return bool(
                self._tab.Get(flatbuffers.number_types.BoolFlags, o + self._tab.Pos)
            )
        return False


def SubscriberFeaturesStart(builder):
    builder.StartObject(8)


def Start(builder):
    SubscriberFeaturesStart(builder)


def SubscriberFeaturesAddPublisherIdentification(builder, publisherIdentification):
    builder.PrependBoolSlot(0, publisherIdentification, 0)


def AddPublisherIdentification(builder, publisherIdentification):
    SubscriberFeaturesAddPublisherIdentification(builder, publisherIdentification)


def SubscriberFeaturesAddPatternBasedSubscription(builder, patternBasedSubscription):
    builder.PrependBoolSlot(1, patternBasedSubscription, 0)


def AddPatternBasedSubscription(builder, patternBasedSubscription):
    SubscriberFeaturesAddPatternBasedSubscription(builder, patternBasedSubscription)


def SubscriberFeaturesAddPublicationTrustlevels(builder, publicationTrustlevels):
    builder.PrependBoolSlot(2, publicationTrustlevels, 0)


def AddPublicationTrustlevels(builder, publicationTrustlevels):
    SubscriberFeaturesAddPublicationTrustlevels(builder, publicationTrustlevels)


def SubscriberFeaturesAddSubscriptionRevocation(builder, subscriptionRevocation):
    builder.PrependBoolSlot(3, subscriptionRevocation, 0)


def AddSubscriptionRevocation(builder, subscriptionRevocation):
    SubscriberFeaturesAddSubscriptionRevocation(builder, subscriptionRevocation)


def SubscriberFeaturesAddEventHistory(builder, eventHistory):
    builder.PrependBoolSlot(4, eventHistory, 0)


def AddEventHistory(builder, eventHistory):
    SubscriberFeaturesAddEventHistory(builder, eventHistory)


def SubscriberFeaturesAddAcknowledgeSubscriberReceived(
    builder, acknowledgeSubscriberReceived
):
    builder.PrependBoolSlot(5, acknowledgeSubscriberReceived, 0)


def AddAcknowledgeSubscriberReceived(builder, acknowledgeSubscriberReceived):
    SubscriberFeaturesAddAcknowledgeSubscriberReceived(
        builder, acknowledgeSubscriberReceived
    )


def SubscriberFeaturesAddPayloadTransparency(builder, payloadTransparency):
    builder.PrependBoolSlot(6, payloadTransparency, 0)


def AddPayloadTransparency(builder, payloadTransparency):
    SubscriberFeaturesAddPayloadTransparency(builder, payloadTransparency)


def SubscriberFeaturesAddPayloadEncryptionCryptobox(
    builder, payloadEncryptionCryptobox
):
    builder.PrependBoolSlot(7, payloadEncryptionCryptobox, 0)


def AddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox):
    SubscriberFeaturesAddPayloadEncryptionCryptobox(builder, payloadEncryptionCryptobox)


def SubscriberFeaturesEnd(builder):
    return builder.EndObject()


def End(builder):
    return SubscriberFeaturesEnd(builder)
